from unittest import mock
import unittest
import os
import shutil
import zipfile
import time
import re
import glob
from webscrapbook import WSB_DIR
from webscrapbook import util
from webscrapbook.scrapbook.host import Host
from webscrapbook.scrapbook.convert import migrate

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_convert')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_input = os.path.join(test_root, 'input')
        cls.test_input_config = os.path.join(cls.test_input, WSB_DIR, 'config.ini')
        cls.test_input_tree = os.path.join(cls.test_input, WSB_DIR, 'tree')
        cls.test_input_meta = os.path.join(cls.test_input_tree, 'meta.js')
        cls.test_input_toc = os.path.join(cls.test_input_tree, 'toc.js')
        cls.test_output = os.path.join(test_root, 'output')
        cls.test_output_tree = os.path.join(cls.test_output, WSB_DIR, 'tree')
        cls.test_output_meta = os.path.join(cls.test_output_tree, 'meta.js')
        cls.test_output_toc = os.path.join(cls.test_output_tree, 'toc.js')

    def setUp(self):
        """Set up a general temp test folder
        """
        os.makedirs(self.test_input_tree, exist_ok=True)
        os.makedirs(self.test_output, exist_ok=True)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_input)
        except NotADirectoryError:
            os.remove(self.test_input)
        except FileNotFoundError:
            pass

        try:
            shutil.rmtree(self.test_output)
        except NotADirectoryError:
            os.remove(self.test_output)
        except FileNotFoundError:
            pass

class TestConvertDataFilesLegacy(Test):
    def test_data_postit(self):
        """Convert postit."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "postit",
    "index": "20200101000000000/index.html"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<html><head><meta http-equiv="Content-Type" content="text/html;Charset=UTF-8"></head><body><pre>
postit page content < & > &lt; &amp; &gt;
</pre></body></html>""")

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """\
<!DOCTYPE html><html><head>\
<meta charset="UTF-8">\
<meta name="viewport" content="width=device-width">\
<style>pre { white-space: pre-wrap; overflow-wrap: break-word; }</style>\
</head><body><pre>
postit page content < & > &lt; &amp; &gt;
</pre></body></html>""")

    def test_data_annotations_linemarker01(self):
        """Convert linemarker."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body><span data-sb-id="1577836800000" data-sb-obj="linemarker" class="linemarker-marked-line" style="background-color: yellow;">Lorem ipsum dolor sit </span><b><span data-sb-id="1577836800000" data-sb-obj="linemarker" class="linemarker-marked-line" style="background-color: yellow;">amet</span></b><span data-sb-id="1577836800000" data-sb-obj="linemarker" class="linemarker-marked-line" style="background-color: yellow;">, consectetur adipiscing elit.</span></body></html>"""

        expected = """<html><body><scrapbook-linemarker data-scrapbook-id="20200101000000000" data-scrapbook-elem="linemarker" class="first" style="background-color: yellow;">Lorem ipsum dolor sit </scrapbook-linemarker><b><scrapbook-linemarker data-scrapbook-id="20200101000000000" data-scrapbook-elem="linemarker" style="background-color: yellow;">amet</scrapbook-linemarker></b><scrapbook-linemarker data-scrapbook-id="20200101000000000" data-scrapbook-elem="linemarker" class="last" style="background-color: yellow;">, consectetur adipiscing elit.</scrapbook-linemarker></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), expected)

    def test_data_annotations_linemarker02(self):
        """Convert legacy ScrapBook linemarker."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body><span class="linemarker-marked-line" style="background-color: yellow;">Lorem ipsum dolor sit </span><b><span class="linemarker-marked-line" style="background-color: yellow;">amet</span></b><span class="linemarker-marked-line" style="background-color: yellow;">, consectetur adipiscing elit.</span></body></html>"""

        expected = """<html><body><scrapbook-linemarker data-scrapbook-elem="linemarker" style="background-color: yellow;">Lorem ipsum dolor sit </scrapbook-linemarker><b><scrapbook-linemarker data-scrapbook-elem="linemarker" style="background-color: yellow;">amet</scrapbook-linemarker></b><scrapbook-linemarker data-scrapbook-elem="linemarker" style="background-color: yellow;">, consectetur adipiscing elit.</scrapbook-linemarker></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), expected)

    def test_data_annotations_inline01(self):
        """Convert inline annotation."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>Lorem ipsum dolor <span data-sb-id="1577836800000" data-sb-obj="inline" class="scrapbook-inline" style="border-bottom: 2px dotted rgb(255, 51, 51); cursor: help;" title="test inline annotation">sit amet, </span><b><span data-sb-id="1577836800000" data-sb-obj="inline" class="scrapbook-inline" style="border-bottom: 2px dotted rgb(255, 51, 51); cursor: help;" title="test inline annotation">consectetur</span></b><span data-sb-id="1577836800000" data-sb-obj="inline" class="scrapbook-inline" style="border-bottom: 2px dotted rgb(255, 51, 51); cursor: help;" title="test inline annotation"> adipiscing elit</span>.</body></html>"""

        expected = """<html><body>Lorem ipsum dolor <scrapbook-linemarker data-scrapbook-id="20200101000000000" data-scrapbook-elem="linemarker" class="first" style="border-bottom: 2px dotted rgb(255, 51, 51); cursor: help;" title="test inline annotation">sit amet, </scrapbook-linemarker><b><scrapbook-linemarker data-scrapbook-id="20200101000000000" data-scrapbook-elem="linemarker" style="border-bottom: 2px dotted rgb(255, 51, 51); cursor: help;" title="test inline annotation">consectetur</scrapbook-linemarker></b><scrapbook-linemarker data-scrapbook-id="20200101000000000" data-scrapbook-elem="linemarker" class="last" style="border-bottom: 2px dotted rgb(255, 51, 51); cursor: help;" title="test inline annotation"> adipiscing elit</scrapbook-linemarker>.<style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_inline02(self):
        """Convert legacy ScrapBook inline annotation."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>Lorem ipsum dolor <span style="border-bottom: 2px dotted #FF3333; cursor: help;" class="scrapbook-inline" title="test inline annotation">sit amet, </span><b><span style="border-bottom: 2px dotted #FF3333; cursor: help;" class="scrapbook-inline" title="123">consectetur</span></b><span style="border-bottom: 2px dotted #FF3333; cursor: help;" class="scrapbook-inline" title="123"> adipiscing elit</span>.</body></html>"""

        expected = """<html><body>Lorem ipsum dolor <scrapbook-linemarker data-scrapbook-elem="linemarker" style="border-bottom: 2px dotted #FF3333; cursor: help;" title="test inline annotation">sit amet, </scrapbook-linemarker><b><scrapbook-linemarker data-scrapbook-elem="linemarker" style="border-bottom: 2px dotted #FF3333; cursor: help;" title="123">consectetur</scrapbook-linemarker></b><scrapbook-linemarker data-scrapbook-elem="linemarker" style="border-bottom: 2px dotted #FF3333; cursor: help;" title="123"> adipiscing elit</scrapbook-linemarker>.<style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_freenote01(self):
        """Convert freenote (absolute)."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body><div data-sb-obj="freenote" style="cursor: help; overflow: visible; margin: 0px; border-width: 12px 1px 1px; border-style: solid; border-color: rgb(204, 204, 204); border-image: none; background: rgb(250, 255, 250) none repeat scroll 0% 0%; opacity: 0.95; padding: 0px; z-index: 500000; text-align: start; font-size: small; line-height: 1.2em; overflow-wrap: break-word; width: 337px; height: 179px; position: absolute; left: 144px; top: 481px;">Test freenote with <b>HTML</b> <u>markups</u>. <br><br>晱瓨扚醏碙螒劦一晹掁舁彾圢柂乜，凗兟兀衩臿趐匼犮屮肸慞垞毌呦。</div></body></html>"""

        expected = """<html><body><scrapbook-sticky data-scrapbook-elem="sticky" class="styled" style="width: 337px; height: 179px; left: 144px; top: 481px;">Test freenote with <b>HTML</b> <u>markups</u>. <br><br>晱瓨扚醏碙螒劦一晹掁舁彾圢柂乜，凗兟兀衩臿趐匼犮屮肸慞垞毌呦。</scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_freenote02(self):
        """Convert freenote (relative)."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body><div data-sb-obj="freenote" style="cursor: help; overflow: visible; margin: 16px auto; border-width: 12px 1px 1px; border-style: solid; border-color: rgb(204, 204, 204); border-image: none; background: rgb(250, 255, 250) none repeat scroll 0% 0%; opacity: 0.95; padding: 0px; z-index: 500000; text-align: start; font-size: small; line-height: 1.2em; overflow-wrap: break-word; width: 233px; height: 89px; position: static;">This is a relative freenote.<br>Anchored between paragraphs.</div></body></html>"""

        expected = """<html><body><scrapbook-sticky data-scrapbook-elem="sticky" class="styled relative" style="width: 233px; height: 89px;">This is a relative freenote.<br>Anchored between paragraphs.</scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_sticky01(self):
        """Convert legacy ScrapBook sticky (absolute)."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><head><link media="all" href="chrome://scrapbook/skin/annotation.css" type="text/css" id="scrapbook-sticky-css" rel="stylesheet"></head><body><div style="left: 376px; top: 107px; position: absolute; width: 240px; height: 79px;" class="scrapbook-sticky"><div class="scrapbook-sticky-header"></div>Sample sticky annotation.
Second line.</div></body></html>"""

        expected = """<html><head></head><body><scrapbook-sticky data-scrapbook-elem="sticky" class="styled plaintext" style="left: 376px; top: 107px; width: 240px; height: 79px;">Sample sticky annotation.
Second line.</scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_sticky02(self):
        """Convert legacy ScrapBook sticky (relative)."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><head><link media="all" href="chrome://scrapbook/skin/annotation.css" type="text/css" id="scrapbook-sticky-css" rel="stylesheet"></head><body><div style="width: 511px; height: 70px;" class="scrapbook-sticky scrapbook-sticky-relative"><div class="scrapbook-sticky-header"></div>Relative sticky.
Second line.</div></body></html>"""

        expected = """<html><head></head><body><scrapbook-sticky data-scrapbook-elem="sticky" class="styled plaintext relative" style="width: 511px; height: 70px;">Relative sticky.
Second line.</scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_sticky03(self):
        """Convert improperly saved legacy ScrapBook sticky (absolute)."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><head><div style="left: 303px; top: 212px; position: absolute; width: 250px; height: 100px;" class="scrapbook-sticky"><div class="scrapbook-sticky-header"></div><textarea style="height: 74px;"></textarea><div class="scrapbook-sticky-footer"><input src="chrome://scrapbook/skin/sticky_save.png" onclick="this.parentNode.parentNode.appendChild(document.createTextNode(this.parentNode.previousSibling.value));this.parentNode.parentNode.removeChild(this.parentNode.previousSibling);this.parentNode.parentNode.removeChild(this.parentNode);" type="image"><input src="chrome://scrapbook/skin/sticky_delete.png" onclick="this.parentNode.parentNode.parentNode.removeChild(this.parentNode.parentNode);" type="image"></div></div></body></html>"""

        expected = """<html><head><scrapbook-sticky data-scrapbook-elem="sticky" class="styled plaintext" style="left: 303px; top: 212px; width: 250px; height: 100px;"></scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_sticky04(self):
        """Convert improperly saved legacy ScrapBook sticky (relative)."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><head><div style="width: 640px; height: 92px;" class="scrapbook-sticky scrapbook-sticky-relative"><div class="scrapbook-sticky-header"></div><textarea style="height: 66px;"></textarea><div class="scrapbook-sticky-footer"><input src="chrome://scrapbook/skin/sticky_save.png" onclick="this.parentNode.parentNode.appendChild(document.createTextNode(this.parentNode.previousSibling.value));this.parentNode.parentNode.removeChild(this.parentNode.previousSibling);this.parentNode.parentNode.removeChild(this.parentNode);" type="image"><input src="chrome://scrapbook/skin/sticky_delete.png" onclick="this.parentNode.parentNode.parentNode.removeChild(this.parentNode.parentNode);" type="image"></div></div></body></html>"""

        expected = """<html><head><scrapbook-sticky data-scrapbook-elem="sticky" class="styled plaintext relative" style="width: 640px; height: 92px;"></scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_block_comment01(self):
        """Convert legacy ScrapBook block comment."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body><div class="scrapbook-block-comment" style="border: 1px dotted rgb(215, 221, 191) !important; margin: 10px !important; padding: 10px !important; font-size: 12px !important; font-weight: normal !important; line-height: 16px !important; text-decoration: none !important; color: rgb(96, 96, 96) !important; background-color: rgb(239, 248, 206) !important; cursor: pointer !important;">This is a sample of a very legacy block comment.</div></body></html>"""

        expected = """<html><body><scrapbook-sticky data-scrapbook-elem="sticky" class="plaintext relative" style="border: 1px dotted rgb(215, 221, 191) !important; margin: 10px !important; padding: 10px !important; font-size: 12px !important; font-weight: normal !important; line-height: 16px !important; text-decoration: none !important; color: rgb(96, 96, 96) !important; background-color: rgb(239, 248, 206) !important; cursor: pointer !important; white-space: pre-wrap;">This is a sample of a very legacy block comment.</scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_block_comment02(self):
        """Convert legacy ScrapBook block comment.

        - Do not add 'white-space: pre-wrap;' style if already exists.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body><div class="scrapbook-block-comment" style="border: 1px dotted rgb(215, 221, 191) !important; margin: 10px !important; padding: 10px !important; font-size: 12px !important; font-weight: normal !important; line-height: 16px !important; text-decoration: none !important; color: rgb(96, 96, 96) !important; background-color: rgb(239, 248, 206) !important; WHITE-SPACE:PRE-WRAP !important; cursor: pointer !important;">This is a sample of a very legacy block comment.</div></body></html>"""

        expected = """<html><body><scrapbook-sticky data-scrapbook-elem="sticky" class="plaintext relative" style="border: 1px dotted rgb(215, 221, 191) !important; margin: 10px !important; padding: 10px !important; font-size: 12px !important; font-weight: normal !important; line-height: 16px !important; text-decoration: none !important; color: rgb(96, 96, 96) !important; background-color: rgb(239, 248, 206) !important; WHITE-SPACE:PRE-WRAP !important; cursor: pointer !important;">This is a sample of a very legacy block comment.</scrapbook-sticky><style data-scrapbook-elem="annotation-css">"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output[:len(expected)], expected)
            self.assertRegex(output, r'<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>')

    def test_data_annotations_other01(self):
        """Convert other legacy ScrapBook elements.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """\
<!DOCTYPE html>
<html>
<head>
<title data-sb-obj="title">My page</title>
</head>
<body>
Donec nec lacus<span data-sb-obj="annotation">(my legacy <em>inline</em>annotation)</span> efficitur.
<a data-sb-obj="link-url" href="http://example.com">Suspendisse eget interdum quam</a>, eu semper <span data-sb-id="1577836800000">ipsum</span>.
</body>
</html>
"""

        expected = """\
<!DOCTYPE html>
<html>
<head>
<title data-scrapbook-elem="title">My page</title>
</head>
<body>
Donec nec lacus<span data-scrapbook-elem="annotation">(my legacy <em>inline</em>annotation)</span> efficitur.
<a data-scrapbook-elem="link-url" href="http://example.com">Suspendisse eget interdum quam</a>, eu semper <span data-scrapbook-id="20200101000000000">ipsum</span>.
</body>
</html>
"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_data_annotations_other02(self):
        """Don't error out if a legacy id is bad.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """\
<!DOCTYPE html>
<html>
<head>
<title data-sb-obj="title">My page</title>
</head>
<body>
<span data-sb-id="">suscipit</span>
<span data-sb-id="999999999999999999999999999999999999999999999999999999999">finibus</span>
<span data-sb-id="abc">varius</span>
</body>
</html>
"""

        expected = """\
<!DOCTYPE html>
<html>
<head>
<title data-scrapbook-elem="title">My page</title>
</head>
<body>
<span data-scrapbook-id="">suscipit</span>
<span data-scrapbook-id="999999999999999999999999999999999999999999999999999999999">finibus</span>
<span data-scrapbook-id="abc">varius</span>
</body>
</html>
"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_data_combine(self):
        """Convert legacy ScrapBook combine page."""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """\
<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<title>Sample combine page</title>
<link rel="stylesheet" href="chrome://scrapbook/skin/combine.css" media="all">
<link rel="stylesheet" href="chrome://scrapbook/skin/annotation.css" media="all">
</head>
<body>

<!--Page1-->
<cite class="scrapbook-header">
	<img src="chrome://scrapbook/skin/treeitem.png" height="16" width="16">
	<a href="http://example.com">Page1</a>
</cite>
<div id="item20200101000000">
<div id="item20200101000000html">
<div id="item20200101000000body">
Page1 content
</div>
</div>
</div>

<!--Page2-->
<cite class="scrapbook-header">
	<img src="chrome://scrapbook/skin/treeitem.png" height="16" width="16">
	<a href="http://example.com">Page1</a>
</cite>
<div id="item20200101000000">
<div id="item20200101000000html">
<div id="item20200101000000body">
Page2 content
</div>
</div>
</div>

<!--Page3-->
<cite class="scrapbook-header">
	<img src="chrome://scrapbook/skin/treenotex.png" height="16" width="16">
	<a href="http://example.com">Page3</a>
</cite>
<div id="item20200102000000">
<div id="item20200102000000html">
<div id="item20200102000000body">
Page3 content
</div>
</div>
</div>
"""

        expected = """\
<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<title>Sample combine page</title>
<style data-scrapbook-elem="custom-css">body { margin: 0px; background-color: #FFFFFF; } cite.scrapbook-header { clear: both; display: block; padding: 3px 6px; font-family: "MS UI Gothic","Tahoma","Verdana","Arial","Sans-Serif","Helvetica"; font-style: normal; font-size: 12px; background-color: InfoBackground; border: 1px solid ThreeDShadow; } cite.scrapbook-header img { vertical-align: middle; } cite.scrapbook-header a { color: InfoText; text-decoration: none; } cite.scrapbook-header a[href]:hover { color: #3388FF; } cite.scrapbook-header a.marked { font-weight: bold; } cite.scrapbook-header a.combine { color: blue; } cite.scrapbook-header a.bookmark { color: limegreen; } cite.scrapbook-header a.notex { color: rgb(80,0,32); } </style>

</head>
<body>

<!--Page1-->
<cite class="scrapbook-header">
\t<img src="treeitem.png" height="16" width="16">
\t<a href="http://example.com">Page1</a>
</cite>
<div id="item20200101000000">
<div id="item20200101000000html">
<div id="item20200101000000body">
Page1 content
</div>
</div>
</div>

<!--Page2-->
<cite class="scrapbook-header">
\t<img src="treeitem.png" height="16" width="16">
\t<a href="http://example.com">Page1</a>
</cite>
<div id="item20200101000000">
<div id="item20200101000000html">
<div id="item20200101000000body">
Page2 content
</div>
</div>
</div>

<!--Page3-->
<cite class="scrapbook-header">
\t<img src="treenotex-1.png" height="16" width="16">
\t<a href="http://example.com">Page3</a>
</cite>
<div id="item20200102000000">
<div id="item20200102000000html">
<div id="item20200102000000body">
Page3 content
</div>
</div>
</div>
"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        with open(os.path.join(self.test_input, '20200101000000000', 'treenotex.png'), 'wb'):
            pass

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000'),
            os.path.join(self.test_output, '20200101000000000', 'index.html'),
            os.path.join(self.test_output, '20200101000000000', 'treeitem.png'),
            os.path.join(self.test_output, '20200101000000000', 'treenotex.png'),
            os.path.join(self.test_output, '20200101000000000', 'treenotex-1.png'),
            })
        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_data_other(self):
        """Convert other legacy ScrapBook attributes.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """\
<!DOCTYPE html>
<html>
<head>
<title>My page</title>
<link rel="stylesheet" href="urn:scrapbook-download-error:https://example.com/style.css">
</head>
<body>
<blockquote><img src="urn:scrapbook-download-skip:https://example.com/picture.jpg"></blockquote>
<blockquote><a data-sb-orig-href="javascript:alert('hello');">script</a></blockquote>
<blockquote><a href="mypage.html" data-sb-indepth="true">in-depth</a></blockquote>
</body>
</html>
"""

        expected = """\
<!DOCTYPE html>
<html>
<head>
<title>My page</title>
<link rel="stylesheet" href="urn:scrapbook-download-error:https://example.com/style.css">
</head>
<body>
<blockquote><img src="urn:scrapbook-download-skip:https://example.com/picture.jpg"></blockquote>
<blockquote><a data-scrapbook-orig-attr-href="javascript:alert('hello');">script</a></blockquote>
<blockquote><a href="mypage.html" data-sb-indepth="true">in-depth</a></blockquote>
</body>
</html>
"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_data_loaders_replace(self):
        """Replace old loaders if there's a change.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>
<span data-sb-obj="custom">foo</span>
<scrapbook-sticky data-scrapbook-elem="sticky" class="styled" style="width: 250px; height: 100px; left: 313px; top: 135px;">abc</scrapbook-sticky>
<style data-scrapbook-elem="annotation-css"></style>
<script data-scrapbook-elem="annotation-loader"></script>
</body></html>"""

        expected_regex = """<html><body>
<span data-scrapbook-elem="custom">foo</span>
<scrapbook-sticky data-scrapbook-elem="sticky" class="styled" style="width: 250px; height: 100px; left: 313px; top: 135px;">abc</scrapbook-sticky>


<style data-scrapbook-elem="annotation-css">(?:[^<]*(?:<(?!/style>)[^<]*)*)</style><script data-scrapbook-elem="annotation-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertRegex(output, expected_regex)

    def test_data_loaders_unchanged(self):
        """Don't update loaders if there's no change detected.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>
<style data-scrapbook-elem="annotation-css"></style>
<script data-scrapbook-elem="annotation-loader"></script>
</body></html>"""

        expected = """<html><body>
<style data-scrapbook-elem="annotation-css"></style>
<script data-scrapbook-elem="annotation-loader"></script>
</body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_data_skip_special_tags(self):
        """Do not rewrite content in <template>, <xml>, <math>, etc.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """\
<!DOCTYPE html>
<html>
<body>
<xmp>
<span data-sb-obj="annotation">foo</span>
</xmp>
<template>
<span data-sb-obj="annotation">foo</span>
</template>
<svg>
<text data-sb-obj="annotation">foo</text>
</svg>
<math>
<mtext data-sb-obj="annotation">foo</mtext>
</math>
</body>
</html>
"""

        expected = """\
<!DOCTYPE html>
<html>
<body>
<xmp>
<span data-sb-obj="annotation">foo</span>
</xmp>
<template>
<span data-sb-obj="annotation">foo</span>
</template>
<svg>
<text data-sb-obj="annotation">foo</text>
</svg>
<math>
<mtext data-sb-obj="annotation">foo</mtext>
</math>
</body>
</html>
"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_legacy=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

class TestConvertDataFilesV1(Test):
    def test_skip_postit(self):
        """Do not touch postit items (even if HTML content happens to meet the criteria).
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "postit",
    "index": "20200101000000000/index.html"
  }
})""")

        input = r"""\
<!DOCTYPE html><html><head>\
<meta charset="UTF-8">\
<meta name="viewport" content="width=device-width">\
<style>pre { white-space: pre-wrap; overflow-wrap: break-word; }</style>\
</head><body><pre>
postit page content
</pre><div data-scrapbook-shadowroot="{&quot;data&quot;:&quot;\n<div>Sub-content.</div>\n\n<p data-scrapbook-shadowroot=\&quot;{&amp;quot;data&amp;quot;:&amp;quot;\\n<div>Deep sub-content.</div>\\n&amp;quot;,&amp;quot;mode&amp;quot;:&amp;quot;open&amp;quot;}\&quot;>Hidden content.</p>&quot;,&quot;mode&quot;:&quot;open&quot;}">Hidden content.</div></body><script data-scrapbook-elem="basic-loader">dummy</script></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, input)

    def test_folder_data_loaders_unchanged(self):
        """Don't update loaders if there's no change detected.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>
<style data-scrapbook-elem="annotation-css"></style>
<script data-scrapbook-elem="annotation-loader"></script>
</body></html>"""

        expected = """<html><body>
<style data-scrapbook-elem="annotation-css"></style>
<script data-scrapbook-elem="annotation-loader"></script>
</body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_folder_data_canvas_loader(self):
        """Migrate [data-scrapbook-elem="canvas-loader"].
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>foo</body><script data-scrapbook-elem="canvas-loader"></script></html>"""

        expected_regex = """<html><body>foo<script data-scrapbook-elem="basic-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertRegex(output, expected_regex)

    def test_folder_data_shadowroot_loader(self):
        """Migrate [data-scrapbook-elem="shadowroot-loader"].
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """<html><body>foo</body><script data-scrapbook-elem="shadowroot-loader"></script></html>"""

        expected_regex = """<html><body>foo<script data-scrapbook-elem="basic-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertRegex(output, expected_regex)

    def test_folder_data_shadowroot(self):
        """Migrate [data-scrapbook-shadowroot].
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = r"""<html><body>foo<div data-scrapbook-shadowroot="{&quot;data&quot;:&quot;\n<div>Sub-content.</div>\n\n<p data-scrapbook-shadowroot=\&quot;{&amp;quot;data&amp;quot;:&amp;quot;\\n<div>Deep sub-content.</div>\\n&amp;quot;,&amp;quot;mode&amp;quot;:&amp;quot;open&amp;quot;}\&quot;>Hidden content.</p>&quot;,&quot;mode&quot;:&quot;open&quot;}">Hidden content.</div></body><script data-scrapbook-elem="basic-loader">dummy</script></html>"""

        expected_regex = """<html><body>foo<div data-scrapbook-shadowdom="
&lt;div&gt;Sub-content.&lt;/div&gt;

&lt;p data-scrapbook-shadowdom=&quot;
&amp;lt;div&amp;gt;Deep sub-content.&amp;lt;/div&amp;gt;
&quot;&gt;Hidden content.&lt;/p&gt;">Hidden content.</div><script data-scrapbook-elem="basic-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertRegex(output, expected_regex)

    def test_folder_data_skip_special_tags(self):
        """Do not rewrite content in <template>, <xml>, <math>, etc.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        input = """\
<!DOCTYPE html>
<html>
<body>
<xmp>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</xmp>
<template>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</template>
<svg>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</svg>
<math>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</math>
</body>
</html>
"""

        expected = """\
<!DOCTYPE html>
<html>
<body>
<xmp>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</xmp>
<template>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</template>
<svg>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</svg>
<math>
<div data-scrapbook-shadowroot='{"data":"foo", "mode": "open"}'></div>
</math>
</body>
</html>
"""

        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000', 'index.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertEqual(output, expected)

    def test_htz_data_shadowroot(self):
        """Migrate [data-scrapbook-shadowroot].
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.htz"
  }
})""")

        input = r"""<html><body>foo<div data-scrapbook-shadowroot="{&quot;data&quot;:&quot;\n<div>Sub-content.</div>\n\n<p data-scrapbook-shadowroot=\&quot;{&amp;quot;data&amp;quot;:&amp;quot;\\n<div>Deep sub-content.</div>\\n&amp;quot;,&amp;quot;mode&amp;quot;:&amp;quot;open&amp;quot;}\&quot;>Hidden content.</p>&quot;,&quot;mode&quot;:&quot;open&quot;}">Hidden content.</div></body><script data-scrapbook-elem="basic-loader">dummy</script></html>"""

        expected_regex = """<html><body>foo<div data-scrapbook-shadowdom="
&lt;div&gt;Sub-content.&lt;/div&gt;

&lt;p data-scrapbook-shadowdom=&quot;
&amp;lt;div&amp;gt;Deep sub-content.&amp;lt;/div&amp;gt;
&quot;&gt;Hidden content.&lt;/p&gt;">Hidden content.</div><script data-scrapbook-elem="basic-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000.htz')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        index_file = os.path.join(self.test_output, '20200101000000000.htz')
        with zipfile.ZipFile(index_file, 'r') as zh:
            output = zh.read('index.html').decode('UTF-8')
            self.assertRegex(output, expected_regex)

    def test_htz_timestamp(self):
        """Don't touch ZIP file if no content is changed.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.htz"
  }
})""")

        input = r"""<html><body>foo</body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000.htz')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', input)

        input_mtime = os.stat(index_file).st_mtime_ns

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        index_file = os.path.join(self.test_output, '20200101000000000.htz')
        with zipfile.ZipFile(index_file, 'r') as zh:
            output = zh.read('index.html').decode('UTF-8')

        output_mtime = os.stat(index_file).st_mtime_ns

        self.assertEqual(output, input)
        self.assertEqual(input_mtime, output_mtime)

    def test_maff_data_shadowroot(self):
        """Migrate [data-scrapbook-shadowroot].
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.maff"
  }
})""")

        input = r"""<html><body>foo<div data-scrapbook-shadowroot="{&quot;data&quot;:&quot;\n<div>Sub-content.</div>\n\n<p data-scrapbook-shadowroot=\&quot;{&amp;quot;data&amp;quot;:&amp;quot;\\n<div>Deep sub-content.</div>\\n&amp;quot;,&amp;quot;mode&amp;quot;:&amp;quot;open&amp;quot;}\&quot;>Hidden content.</p>&quot;,&quot;mode&quot;:&quot;open&quot;}">Hidden content.</div></body><script data-scrapbook-elem="basic-loader">dummy</script></html>"""

        expected_regex = """<html><body>foo<div data-scrapbook-shadowdom="
&lt;div&gt;Sub-content.&lt;/div&gt;

&lt;p data-scrapbook-shadowdom=&quot;
&amp;lt;div&amp;gt;Deep sub-content.&amp;lt;/div&amp;gt;
&quot;&gt;Hidden content.&lt;/p&gt;">Hidden content.</div><script data-scrapbook-elem="basic-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000.maff')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', input)
            zh.writestr('20200101000000000/index.rdf', """dummy""")

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        index_file = os.path.join(self.test_output, '20200101000000000.maff')
        with zipfile.ZipFile(index_file, 'r') as zh:
            output = zh.read('20200101000000000/index.html').decode('UTF-8')
            self.assertRegex(output, expected_regex)

    def test_maff_timestamp(self):
        """Don't touch ZIP file if no content is changed.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.maff"
  }
})""")

        input = r"""<html><body>foo</body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000.maff')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', input)
            zh.writestr('20200101000000000/index.rdf', """dummy""")

        input_mtime = os.stat(index_file).st_mtime_ns

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        index_file = os.path.join(self.test_output, '20200101000000000.maff')
        with zipfile.ZipFile(index_file, 'r') as zh:
            output = zh.read('20200101000000000/index.html').decode('UTF-8')

        output_mtime = os.stat(index_file).st_mtime_ns

        self.assertEqual(output, input)
        self.assertEqual(input_mtime, output_mtime)

    def test_single_file_data_shadowroot(self):
        """Migrate [data-scrapbook-shadowroot].
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.html"
  }
})""")

        input = r"""<html><body>foo<div data-scrapbook-shadowroot="{&quot;data&quot;:&quot;\n<div>Sub-content.</div>\n\n<p data-scrapbook-shadowroot=\&quot;{&amp;quot;data&amp;quot;:&amp;quot;\\n<div>Deep sub-content.</div>\\n&amp;quot;,&amp;quot;mode&amp;quot;:&amp;quot;open&amp;quot;}\&quot;>Hidden content.</p>&quot;,&quot;mode&quot;:&quot;open&quot;}">Hidden content.</div></body><script data-scrapbook-elem="basic-loader">dummy</script></html>"""

        expected_regex = """<html><body>foo<div data-scrapbook-shadowdom="
&lt;div&gt;Sub-content.&lt;/div&gt;

&lt;p data-scrapbook-shadowdom=&quot;
&amp;lt;div&amp;gt;Deep sub-content.&amp;lt;/div&amp;gt;
&quot;&gt;Hidden content.&lt;/p&gt;">Hidden content.</div><script data-scrapbook-elem="basic-loader">(?:[^<]*(?:<(?!/script>)[^<]*)*)</script></body></html>"""

        index_file = os.path.join(self.test_input, '20200101000000000.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write(input)

        for info in migrate.run(self.test_input, self.test_output, convert_v1=True):
            pass

        with open(os.path.join(self.test_output, '20200101000000000.html'), encoding='UTF-8') as fh:
            output = fh.read()
            self.assertRegex(output, expected_regex)

if __name__ == '__main__':
    unittest.main()



from unittest import mock
import unittest
import os
import shutil
import zipfile
import time
import re
import glob
from base64 import b64decode
from email.utils import format_datetime

from webscrapbook import WSB_DIR
from webscrapbook import util
from webscrapbook.scrapbook.host import Host
from webscrapbook.scrapbook.convert import items as conv_items

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_convert')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestRun(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_input = os.path.join(test_root, 'input')
        cls.test_input_config = os.path.join(cls.test_input, WSB_DIR, 'config.ini')
        cls.test_input_tree = os.path.join(cls.test_input, WSB_DIR, 'tree')
        cls.test_input_meta = os.path.join(cls.test_input_tree, 'meta.js')
        cls.test_input_toc = os.path.join(cls.test_input_tree, 'toc.js')
        cls.test_output = os.path.join(test_root, 'output')
        cls.test_output_tree = os.path.join(cls.test_output, WSB_DIR, 'tree')
        cls.test_output_meta = os.path.join(cls.test_output_tree, 'meta.js')
        cls.test_output_toc = os.path.join(cls.test_output_tree, 'toc.js')

    def setUp(self):
        """Set up a general temp test folder
        """
        os.makedirs(self.test_input_tree, exist_ok=True)
        os.makedirs(self.test_output, exist_ok=True)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_input)
        except NotADirectoryError:
            os.remove(self.test_input)
        except FileNotFoundError:
            pass

        try:
            shutil.rmtree(self.test_output)
        except NotADirectoryError:
            os.remove(self.test_output)
        except FileNotFoundError:
            pass

    def test_param_type(self):
        """Check type filter
        """
        types = ['', 'site', 'image', 'file', 'combine', 'note', 'postit', 'bookmark', 'folder', 'separator']
        for type in types:
            with self.subTest(type=type):
                self.setUp()
                try:
                    with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
                        fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "%s",
    "index": "20200101000000000/index.html"
  }
})""" % type)

                    index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
                    os.makedirs(os.path.dirname(index_file), exist_ok=True)
                    with open(index_file, 'w', encoding='UTF-8') as fh:
                        fh.write("""dummy""")

                    for info in conv_items.run(self.test_input, self.test_output, format='htz', types=[type]):
                        pass

                    self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
                        os.path.join(self.test_output, ''),
                        os.path.join(self.test_output, '20200101000000000.htz'),
                        })
                finally:
                    self.tearDown()

                self.setUp()
                try:
                    with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
                        fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "%s",
    "index": "20200101000000000/index.html"
  }
})""" % type)

                    index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
                    os.makedirs(os.path.dirname(index_file), exist_ok=True)
                    with open(index_file, 'w', encoding='UTF-8') as fh:
                        fh.write("""dummy""")

                    for info in conv_items.run(self.test_input, self.test_output, format='htz', types=[]):
                        pass

                    self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
                        os.path.join(self.test_output, ''),
                        os.path.join(self.test_output, '20200101000000000'),
                        os.path.join(self.test_output, '20200101000000000', 'index.html'),
                        })
                finally:
                    self.tearDown()

    def _test_param_format_sample(self):
        """Generate sample files for test_param_format_*
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "create": "20200101000000000",
    "index": "20200101000000001/index.html"
  },
  "20200101000000002": {
    "type": "",
    "create": "20200101000000000",
    "index": "20200101000000002.htz"
  },
  "20200101000000003": {
    "type": "",
    "create": "20200101000000000",
    "index": "20200101000000003.maff"
  },
  "20200101000000004": {
    "type": "",
    "create": "20200101000000000",
    "index": "20200101000000004.html"
  },
  "20200101000000005": {
    "type": "",
    "create": "20200101000000000",
    "index": "20200101000000005.txt"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000001')
        os.makedirs(index_dir, exist_ok=True)
        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""\
<img src="resource.bmp">
my page content
""")
        with open(os.path.join(index_dir, 'resource.bmp'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        index_file = os.path.join(self.test_input, '20200101000000002.htz')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', """\
<img src="resource.bmp">
my page content
""")
            zh.writestr('resource.bmp', b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        index_file = os.path.join(self.test_input, '20200101000000003.maff')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000003/index.html', """\
<img src="resource.bmp">
my page content
""")
            zh.writestr('20200101000000003/index.rdf', """dummy""")
            zh.writestr('20200101000000003/resource.bmp', b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        index_file = os.path.join(self.test_input, '20200101000000004.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<img src="data:image/bmp;base64,Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA">
my page content
""")

        index_file = os.path.join(self.test_input, '20200101000000005.txt')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")

    def test_param_format_folder01(self):
        """Test format "folder"

        - Folder converted from MAFF should not contain index.rdf
        """
        self._test_param_format_sample()

        for info in conv_items.run(self.test_input, self.test_output, format='folder', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000001'),
            os.path.join(self.test_output, '20200101000000001', 'index.html'),
            os.path.join(self.test_output, '20200101000000001', 'resource.bmp'),
            os.path.join(self.test_output, '20200101000000002'),
            os.path.join(self.test_output, '20200101000000002', 'index.html'),
            os.path.join(self.test_output, '20200101000000002', 'resource.bmp'),
            os.path.join(self.test_output, '20200101000000003'),
            os.path.join(self.test_output, '20200101000000003', 'index.html'),
            os.path.join(self.test_output, '20200101000000003', 'resource.bmp'),
            os.path.join(self.test_output, '20200101000000004'),
            os.path.join(self.test_output, '20200101000000004', 'index.html'),
            os.path.join(self.test_output, '20200101000000004', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp'),
            os.path.join(self.test_output, '20200101000000005'),
            os.path.join(self.test_output, '20200101000000005', 'index.html'),
            os.path.join(self.test_output, '20200101000000005', '20200101000000005.txt'),
            })

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000001', 'index.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000001', 'index.html')).st_mtime,
            )

        with zipfile.ZipFile(os.path.join(self.test_input, '20200101000000002.htz')) as zh:
            self.assertEqual(
                util.zip_timestamp(zh.getinfo('index.html')),
                os.stat(os.path.join(self.test_output, '20200101000000002', 'index.html')).st_mtime,
                )

        with zipfile.ZipFile(os.path.join(self.test_input, '20200101000000003.maff')) as zh:
            self.assertEqual(
                util.zip_timestamp(zh.getinfo('20200101000000003/index.html')),
                os.stat(os.path.join(self.test_output, '20200101000000003', 'index.html')).st_mtime,
                )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000004.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000004', 'index.html')).st_mtime,
            )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000005.txt')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000005', 'index.html')).st_mtime,
            )

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000001/index.html',
                },
            '20200101000000002': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000002/index.html',
                },
            '20200101000000003': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000003/index.html',
                },
            '20200101000000004': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000004/index.html',
                },
            '20200101000000005': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000005/index.html',
                },
            })

    def test_param_format_folder02(self):
        """Check if icon path is correctly handled for htz => folder.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.htz",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.htz')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', """my page content""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='folder', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000'),
            os.path.join(self.test_output, '20200101000000000', 'index.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000/index.html',
                'icon': '../.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_folder03(self):
        """Check if icon path is correctly handled for maff => folder.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.maff",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.maff')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', """my page content""")
            zh.writestr('20200101000000000/index.rdf', """dummy""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='folder', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000'),
            os.path.join(self.test_output, '20200101000000000', 'index.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000/index.html',
                'icon': '../.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_folder04(self):
        """Check if icon path is correctly handled for single_file => folder.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.html",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='folder', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000'),
            os.path.join(self.test_output, '20200101000000000', 'index.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000/index.html',
                'icon': '../.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_htz01(self):
        """Test format "htz"

        - HTZ converted from MAFF should not contain index.rdf
        """
        self._test_param_format_sample()

        for info in conv_items.run(self.test_input, self.test_output, format='htz', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000001.htz'),
            os.path.join(self.test_output, '20200101000000002.htz'),
            os.path.join(self.test_output, '20200101000000003.htz'),
            os.path.join(self.test_output, '20200101000000004.htz'),
            os.path.join(self.test_output, '20200101000000005.htz'),
            })

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000001.htz')) as zh:
            self.assertEqual(set(zh.namelist()), {'index.html', 'resource.bmp'})

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000003.htz')) as zh:
            self.assertEqual(set(zh.namelist()), {'index.html', 'resource.bmp'})

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000004.htz')) as zh:
            self.assertEqual(set(zh.namelist()), {'index.html', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp'})

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000005.htz')) as zh:
            self.assertEqual(set(zh.namelist()), {'index.html', '20200101000000005.txt'})

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000001', 'index.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000001.htz')).st_mtime,
            )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000002.htz')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000002.htz')).st_mtime,
            )

        with zipfile.ZipFile(os.path.join(self.test_input, '20200101000000003.maff')) as zh:
            self.assertEqual(
                util.zip_timestamp(zh.getinfo('20200101000000003/index.html')),
                os.stat(os.path.join(self.test_output, '20200101000000003.htz')).st_mtime,
                )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000004.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000004.htz')).st_mtime,
            )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000005.txt')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000005.htz')).st_mtime,
            )

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000001.htz',
                },
            '20200101000000002': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000002.htz',
                },
            '20200101000000003': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000003.htz',
                },
            '20200101000000004': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000004.htz',
                },
            '20200101000000005': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000005.htz',
                },
            })

    def test_param_format_htz02(self):
        """Check if icon path is correctly handled for folder => htz.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html",
    "icon": "favicon.ico"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)
        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")
        with open(os.path.join(index_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='htz', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.htz'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.htz',
                'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.ico',
                },
            })

    def test_param_format_htz03(self):
        """Check if icon path is correctly handled for maff => htz.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.maff",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.maff')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', """my page content""")
            zh.writestr('20200101000000000/index.rdf', """dummy""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='htz', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.htz'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.htz',
                'icon': '.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_htz04(self):
        """Check if icon path is correctly handled for single_file => htz.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.html",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'icon': '.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_maff01(self):
        """Test format "maff"

        - MAFF converted from other format should contain a valid index.rdf
        """
        self._test_param_format_sample()

        for info in conv_items.run(self.test_input, self.test_output, format='maff', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000001.maff'),
            os.path.join(self.test_output, '20200101000000002.maff'),
            os.path.join(self.test_output, '20200101000000003.maff'),
            os.path.join(self.test_output, '20200101000000004.maff'),
            os.path.join(self.test_output, '20200101000000005.maff'),
            })

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000001.maff')) as zh:
            self.assertEqual(set(zh.namelist()), {
                '20200101000000001/',
                '20200101000000001/index.html',
                '20200101000000001/index.rdf',
                '20200101000000001/resource.bmp',
                })
            with zh.open('20200101000000001/index.rdf') as fh:
                dt = util.id_to_datetime('20200101000000000').astimezone()
                self.assertEqual(
                    util.parse_maff_index_rdf(fh),
                    ('', '', format_datetime(dt), 'index.html', 'UTF-8'),
                    )

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000002.maff')) as zh:
            self.assertEqual(set(zh.namelist()), {
                '20200101000000002/',
                '20200101000000002/index.html',
                '20200101000000002/index.rdf',
                '20200101000000002/resource.bmp',
                })
            with zh.open('20200101000000002/index.rdf') as fh:
                dt = util.id_to_datetime('20200101000000000').astimezone()
                self.assertEqual(
                    util.parse_maff_index_rdf(fh),
                    ('', '', format_datetime(dt), 'index.html', 'UTF-8'),
                    )

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000004.maff')) as zh:
            self.assertEqual(set(zh.namelist()), {
                '20200101000000004/',
                '20200101000000004/index.html',
                '20200101000000004/index.rdf',
                '20200101000000004/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp',
                })
            with zh.open('20200101000000004/index.rdf') as fh:
                dt = util.id_to_datetime('20200101000000000').astimezone()
                self.assertEqual(
                    util.parse_maff_index_rdf(fh),
                    ('', '', format_datetime(dt), 'index.html', 'UTF-8'),
                    )

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000005.maff')) as zh:
            self.assertEqual(set(zh.namelist()), {
                '20200101000000005/',
                '20200101000000005/index.html',
                '20200101000000005/index.rdf',
                '20200101000000005/20200101000000005.txt',
                })
            with zh.open('20200101000000005/index.rdf') as fh:
                dt = util.id_to_datetime('20200101000000000').astimezone()
                self.assertEqual(
                    util.parse_maff_index_rdf(fh),
                    ('', '', format_datetime(dt), 'index.html', 'UTF-8'),
                    )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000001', 'index.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000001.maff')).st_mtime,
            )

        with zipfile.ZipFile(os.path.join(self.test_input, '20200101000000002.htz')) as zh:
            self.assertEqual(
                util.zip_timestamp(zh.getinfo('index.html')),
                os.stat(os.path.join(self.test_output, '20200101000000002.maff')).st_mtime,
                )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000003.maff')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000003.maff')).st_mtime,
            )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000004.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000004.maff')).st_mtime,
            )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000005.txt')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000005.maff')).st_mtime,
            )

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001.maff',
                'create': '20200101000000000',
                },
            '20200101000000002': {
                'type': '',
                'index': '20200101000000002.maff',
                'create': '20200101000000000',
                },
            '20200101000000003': {
                'type': '',
                'index': '20200101000000003.maff',
                'create': '20200101000000000',
                },
            '20200101000000004': {
                'type': '',
                'index': '20200101000000004.maff',
                'create': '20200101000000000',
                },
            '20200101000000005': {
                'type': '',
                'index': '20200101000000005.maff',
                'create': '20200101000000000',
                },
            })

    def test_param_format_maff02(self):
        """Fail if index.rdf already exists
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "index": "20200101000000001/index.html"
  },
  "20200101000000002": {
    "type": "",
    "index": "20200101000000002.htz"
  },
  "20200101000000003": {
    "type": "",
    "index": "20200101000000003.maff"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000001')
        os.makedirs(index_dir, exist_ok=True)
        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")
        with open(os.path.join(index_dir, 'index.rdf'), 'w', encoding='UTF-8') as fh:
            fh.write("""dummy""")
        with open(os.path.join(index_dir, 'resource.txt'), 'w', encoding='UTF-8') as fh:
            fh.write("""dummy""")

        index_file = os.path.join(self.test_input, '20200101000000002.htz')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', """my page content""")
            zh.writestr('index.rdf', """dummy""")
            zh.writestr('resource.txt', """dummy""")

        index_file = os.path.join(self.test_input, '20200101000000003.maff')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000003/index.html', """my page content""")
            zh.writestr('20200101000000003/index.rdf', """dummy""")
            zh.writestr('20200101000000003/resource.txt', """dummy""")

        for info in conv_items.run(self.test_input, self.test_output, format='maff', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000001'),
            os.path.join(self.test_output, '20200101000000001', 'index.html'),
            os.path.join(self.test_output, '20200101000000001', 'index.rdf'),
            os.path.join(self.test_output, '20200101000000001', 'resource.txt'),
            os.path.join(self.test_output, '20200101000000002.htz'),
            os.path.join(self.test_output, '20200101000000003.maff'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                },
            '20200101000000002': {
                'type': '',
                'index': '20200101000000002.htz',
                },
            '20200101000000003': {
                'type': '',
                'index': '20200101000000003.maff',
                },
            })

    def test_param_format_maff03(self):
        """Check if icon path is correctly handled for folder => maff.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html",
    "icon": "favicon.ico"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)
        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")
        with open(os.path.join(index_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='maff', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.maff'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.maff',
                'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.ico',
                },
            })

    def test_param_format_maff04(self):
        """Check if icon path is correctly handled for htz => maff.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.htz",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.htz')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', """my page content""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='maff', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.maff'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.maff',
                'icon': '.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_maff05(self):
        """Check if icon path is correctly handled for single_file => maff.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.html",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='maff', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.maff'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.maff',
                'icon': '.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_single_file01(self):
        """Test format "single_file"
        """
        self._test_param_format_sample()

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000001.html'),
            os.path.join(self.test_output, '20200101000000002.html'),
            os.path.join(self.test_output, '20200101000000003.html'),
            os.path.join(self.test_output, '20200101000000004.html'),
            os.path.join(self.test_output, '20200101000000005.txt'),
            })

        with open( os.path.join(self.test_output, '20200101000000001.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """\
<img src="data:image/bmp;base64,Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA">
my page content
""")

        with open( os.path.join(self.test_output, '20200101000000002.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """\
<img src="data:image/bmp;base64,Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA">
my page content
""")

        with open( os.path.join(self.test_output, '20200101000000003.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """\
<img src="data:image/bmp;base64,Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA">
my page content
""")

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000001', 'index.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000001.html')).st_mtime,
            )

        with zipfile.ZipFile(os.path.join(self.test_input, '20200101000000002.htz')) as zh:
            self.assertEqual(
                util.zip_timestamp(zh.getinfo('index.html')),
                os.stat(os.path.join(self.test_output, '20200101000000002.html')).st_mtime,
                )

        with zipfile.ZipFile(os.path.join(self.test_input, '20200101000000003.maff')) as zh:
            self.assertEqual(
                util.zip_timestamp(zh.getinfo('20200101000000003/index.html')),
                os.stat(os.path.join(self.test_output, '20200101000000003.html')).st_mtime,
                )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000004.html')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000004.html')).st_mtime,
            )

        self.assertEqual(
            os.stat(os.path.join(self.test_input, '20200101000000005.txt')).st_mtime,
            os.stat(os.path.join(self.test_output, '20200101000000005.txt')).st_mtime,
            )

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000001.html',
                },
            '20200101000000002': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000002.html',
                },
            '20200101000000003': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000003.html',
                },
            '20200101000000004': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000004.html',
                },
            '20200101000000005': {
                'type': '',
                'create': '20200101000000000',
                'index': '20200101000000005.txt',
                },
            })

    def test_param_format_single_file02(self):
        """Check if icon path is correctly handled for folder => single_file.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html",
    "icon": "favicon.ico"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)
        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""my page content""")
        with open(os.path.join(index_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.ico',
                },
            })

    def test_param_format_single_file03(self):
        """Check if icon path is correctly handled for htz => single_file.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.htz",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.htz')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', """my page content""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'icon': '.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_single_file04(self):
        """Check if icon path is correctly handled for maff => single_file.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000.maff",
    "icon": ".wsb/tree/favicon/favicon.ico"
  }
})""")

        index_file = os.path.join(self.test_input, '20200101000000000.maff')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', """my page content""")
            zh.writestr('20200101000000000/index.rdf', """dummy""")

        favicon_dir = os.path.join(self.test_input_tree, 'favicon')
        os.makedirs(favicon_dir, exist_ok=True)
        with open(os.path.join(favicon_dir, 'favicon.ico'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', '**'), recursive=True)), {
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', ''),
            os.path.join(self.test_output, WSB_DIR, 'tree', 'favicon', 'favicon.ico'),
            })

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'icon': '.wsb/tree/favicon/favicon.ico',
                },
            })

    def test_param_format_single_file05(self):
        """Check if meta refresh is resolved recursively.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)

        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""<meta http-equiv="refresh" content="0; url=./refresh1.html">""")

        with open(os.path.join(index_dir, 'refresh1.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""<meta http-equiv="refresh" content="0; url=./refresh2.html">""")

        with open(os.path.join(index_dir, 'refresh2.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""page content""")

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        with open(os.path.join(self.test_output, '20200101000000000.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """page content""")

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                },
            })

    def test_param_format_single_file06(self):
        """Check if meta refresh target is non-HTML, and SVG rewriting.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)

        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""<meta http-equiv="refresh" content="0; url=./target.svg">""")

        with open(os.path.join(index_dir, 'target.svg'), 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg xmlns="http://www.w3.org/2000/svg">
  <image href="./image.bmp"/>
</svg>
""")

        with open(os.path.join(index_dir, 'image.bmp'), 'wb') as fh:
            fh.write(b64decode(b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.svg'),
            })

        with open(os.path.join(self.test_output, '20200101000000000.svg'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """\
<?xml version="1.0"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg xmlns="http://www.w3.org/2000/svg">
  <image href="data:image/bmp;base64,Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA" />
</svg>
""")

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.svg',
                },
            })

    def test_param_format_single_file07(self):
        """Check that a deleyed meta refresh should not be resolved.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)

        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""<meta http-equiv="refresh" content="1; url=./target.html">""")

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        with open(os.path.join(self.test_output, '20200101000000000.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """<meta http-equiv="refresh" content="1; url=urn:scrapbook:convert:skip:url:./target.html">""")

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                },
            })

    def test_param_format_single_file08(self):
        """Check that a meta refresh to an absolute URL should not be resolved.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "index": "20200101000000000/index.html"
  }
})""")

        index_dir = os.path.join(self.test_input, '20200101000000000')
        os.makedirs(index_dir, exist_ok=True)

        with open(os.path.join(index_dir, 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write("""<meta http-equiv="refresh" content="0; url=http://example.com">""")

        for info in conv_items.run(self.test_input, self.test_output, format='single_file', types=['']):
            pass

        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, '20200101000000000.html'),
            })

        with open(os.path.join(self.test_output, '20200101000000000.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """<meta http-equiv="refresh" content="0; url=http://example.com">""")

        book = Host(self.test_output).books['']
        book.load_meta_files()
        self.assertDictEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                },
            })

if __name__ == '__main__':
    unittest.main()



from unittest import mock
import unittest
import os
from webscrapbook.util.html import Markup, MarkupTag, HtmlRewriter

root_dir = os.path.abspath(os.path.dirname(__file__))

class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192

class TestMarkupTag(Test):
    def test_html01(self):
        m = MarkupTag(
            type='starttag',
            tag='input',
            attrs=[
                ('type', 'checkbox'),
                ('checked', None),
                ],
            )
        self.assertEqual(str(m), '<input type="checkbox" checked>')

    def test_html02(self):
        m = MarkupTag(
            type='starttag',
            tag='div',
            attrs=[
                ('title', '中文<span title="foo">bar</span>'),
                ],
            )
        self.assertEqual(str(m), '<div title="中文&lt;span title=&quot;foo&quot;&gt;bar&lt;/span&gt;">')

    def test_xhtml01(self):
        m = MarkupTag(
            is_xhtml=True,
            type='starttag',
            tag='input',
            attrs=[
                ('type', 'checkbox'),
                ('checked', None),
                ],
            is_self_end=True,
            )
        self.assertEqual(str(m), '<input type="checkbox" checked="checked" />')

    def test_xhtml02(self):
        m = MarkupTag(
            is_xhtml=True,
            type='starttag',
            tag='div',
            attrs=[
                ('title', '中文<span title="foo">bar</span>'),
                ],
            )
        self.assertEqual(str(m), '<div title="中文&lt;span title=&quot;foo&quot;&gt;bar&lt;/span&gt;">')

class TestHtmlRewriter(Test):
    def test_loads_html01(self):
        input = """<!DOCTYPE html>"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html02(self):
        input = """<meta charset="UTF-8">"""
        parsed = """<meta charset="UTF-8"></meta>"""
        reserialized = """<meta charset="UTF-8">"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html03(self):
        input = """\
<style>
<!--/*--><![CDATA[/*><!--*/
body::after { content: "<my> <" "/style> & godness"; }
/*]]>*/-->
</style>
"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html04(self):
        input = """\
<script>
<!--//--><![CDATA[//><!--
console.log('test <my> <' + '/script> & tag');
//--><!]]>
</script>
"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html05(self):
        input = """\
foo&nbsp;&nbsp;&nbsp;中文<br>
King&apos;s &quot;123&quot; &lt; &amp; &gt; 456 (escaped)<br>
King's "123" < & > 456 (unescaped)<br>
"""
        parsed = """\
foo&nbsp;&nbsp;&nbsp;中文<br></br>
King&apos;s &quot;123&quot; &lt; &amp; &gt; 456 (escaped)<br></br>
King's "123" < & > 456 (unescaped)<br></br>
"""
        reserialized = """\
foo&nbsp;&nbsp;&nbsp;中文<br>
King&apos;s &quot;123&quot; &lt; &amp; &gt; 456 (escaped)<br>
King's "123" < & > 456 (unescaped)<br>
"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html06(self):
        input = """<SPaN TITLE="mytitle">Tag with mixed case</SpAN>"""
        parsed = input
        reserialized = """<span title="mytitle">Tag with mixed case</span>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html07(self):
        input = """<input type="checkbox" checked>"""
        parsed = """<input type="checkbox" checked></input>"""
        reserialized = """<input type="checkbox" checked>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html08(self):
        input = """<input type="number"/>"""
        parsed = """<input type="number"/></input>"""
        reserialized = """<input type="number" />"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html09(self):
        input = """<div/>non-void-html-element-start-tag-with-trailing-solidus</div>"""
        parsed = input
        reserialized = """<div />non-void-html-element-start-tag-with-trailing-solidus</div>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html10(self):
        input = """<textarea>plain <div> text may be escaped &lt; &amp; &gt;</textarea>"""
        parsed = """<textarea>plain <div> text may be escaped &lt; &amp; &gt;</textarea>"""
        reserialized = parsed

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html11(self):
        input = """<!UNKONWN decl>"""
        parsed = input
        reserialized = """<!--UNKONWN decl-->"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html12(self):
        input = """<![CDATA[x<y]]>"""
        parsed = input
        reserialized = """<!CDATA[x<y>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html13(self):
        input = """<![if sth]>"""
        parsed = input
        reserialized = """<!if sth>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html14(self):
        input = """<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"/><text x="25" y="40">demo</text></svg>"""
        parsed = """<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"/></circle><text x="25" y="40">demo</text></svg>"""
        reserialized = """<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow" /><text x="25" y="40">demo</text></svg>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html15(self):
        input = """<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"><text x="25" y="40">demo</text></svg>"""
        parsed = """<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"><text x="25" y="40">demo</text></circle></svg>"""
        reserialized = """<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"><text x="25" y="40">demo</text></svg>"""

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_html16(self):
        input = """\
<math>
  <mrow>
    <mi>x</mi>
    <mo>=</mo>
    <mfrac>
      <mrow>
        <mo form="prefix">&#x2212;<!-- − --></mo>
        <mi>b</mi>
        <mo>&#x00B1;<!-- &PlusMinus; --></mo>
        <msqrt>
          <msup>
            <mi>b</mi>
            <mn>2</mn>
          </msup>
          <mo>&#x2212;<!-- − --></mo>
          <mn>4</mn>
          <mo>&#x2062;<!-- &InvisibleTimes; --></mo>
          <mi>a</mi>
          <mo>&#x2062;<!-- &InvisibleTimes; --></mo>
          <mi>c</mi>
        </msqrt>
      </mrow>
      <mrow>
        <mn>2</mn>
        <mo>&#x2062;<!-- &InvisibleTimes; --></mo>
        <mi>a</mi>
      </mrow>
    </mfrac>
  </mrow>
</math>
"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter().loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml01(self):
        input = """<?xml version="1.0" encoding="UTF-8"?>"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml02(self):
        input = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml03(self):
        input = """<meta charset="UTF-8"/>"""
        parsed = """<meta charset="UTF-8"/></meta>"""
        reserialized = """<meta charset="UTF-8" />"""

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml04(self):
        input = """\
<style>
<!--/*--><![CDATA[/*><!--*/
body::after { content: "<my> <" "/style> & godness"; }
/*]]>*/-->
</style>
"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml05(self):
        input = """\
<script type="text/javascript">
<!--//--><![CDATA[//><!--
console.log('test <my> <' + '/script> & tag');
//--><!]]>
</script>
"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml06(self):
        input = """\
foo&nbsp;&nbsp;&nbsp;中文<br/>
King&apos;s &quot;123&quot; &lt; &amp; &gt; 456 (escaped)<br/>
"""
        parsed = """\
foo&nbsp;&nbsp;&nbsp;中文<br/></br>
King&apos;s &quot;123&quot; &lt; &amp; &gt; 456 (escaped)<br/></br>
"""
        reserialized = """\
foo&nbsp;&nbsp;&nbsp;中文<br />
King&apos;s &quot;123&quot; &lt; &amp; &gt; 456 (escaped)<br />
"""

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml07(self):
        input = """<input type="checkbox" checked="checked"/>"""
        parsed = """<input type="checkbox" checked="checked"/></input>"""
        reserialized = """<input type="checkbox" checked="checked" />"""

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml08(self):
        input = """<input type="checkbox" checked="checked"></input>"""
        parsed = input
        reserialized = """<input type="checkbox" checked="checked"></input>"""

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml09(self):
        input = """<textarea>text need escaped &lt; &amp; &gt;</textarea>"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml10(self):
        input = """<svg xmlns="http://www.w3.org/2000/svg"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"/><text x="25" y="40">demo</text></svg>"""
        parsed = """<svg xmlns="http://www.w3.org/2000/svg"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"/></circle><text x="25" y="40">demo</text></svg>"""
        reserialized = """<svg xmlns="http://www.w3.org/2000/svg"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow" /><text x="25" y="40">demo</text></svg>"""

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)

    def test_loads_xhtml11(self):
        input = """\
<math xmlns="http://www.w3.org/1998/Math/MathML">
  <mrow>
    <mi>x</mi>
    <mo>=</mo>
    <mfrac>
      <mrow>
        <mo form="prefix">&#x2212;<!-- − --></mo>
        <mi>b</mi>
        <mo>&#x00B1;<!-- &PlusMinus; --></mo>
        <msqrt>
          <msup>
            <mi>b</mi>
            <mn>2</mn>
          </msup>
          <mo>&#x2212;<!-- − --></mo>
          <mn>4</mn>
          <mo>&#x2062;<!-- &InvisibleTimes; --></mo>
          <mi>a</mi>
          <mo>&#x2062;<!-- &InvisibleTimes; --></mo>
          <mi>c</mi>
        </msqrt>
      </mrow>
      <mrow>
        <mn>2</mn>
        <mo>&#x2062;<!-- &InvisibleTimes; --></mo>
        <mi>a</mi>
      </mrow>
    </mfrac>
  </mrow>
</math>
"""
        parsed = input
        reserialized = parsed

        markups = HtmlRewriter(is_xhtml=True).loads(input)
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), input)
        self.assertEqual(''.join(str(m) for m in markups), parsed)
        for m in markups: m.src = None
        self.assertEqual(''.join(str(m) for m in markups if not m.hidden), reserialized)


if __name__ == '__main__':
    unittest.main()
